#!/usr/bin/env python3
import os
from os.path import expandvars
from subprocess import CompletedProcess
from lmn.helpers import posixpath2str, replace_lmn_envvars
from typing import Optional
import invoke

from lmn import logger

LMN_DOCKER_ROOTDIR = '/lmn'

# NOTE: Should I have ssh-conf, slurm-conf and docker-conf separately??
# I guess RemoteConfig should ONLY store the info on how to login to the host?
# docker info and slurm info should really reside in project.
class RemoteConfig:
    """Represents a configuration to connect to a remote server.
    This is used by SimpleSSHClient.
    """
    def __init__(self, user, host, port=22) -> None:
        self.user = user
        self.host = host
        self._conn = None
        # self.port = port

    @property
    def base_uri(self) -> str:
        return f'{self.user}@{self.host}'

    # from fabric import Connection
    # def get_connection(self) -> Connection:
    #     from fabric import Connection
    #     from fabric.config import Config
    #     config = Config()
    #     config.user = self.user
    #     conn = Connection(host=self.host, config=config, inline_ssh_env=True)
    #     conn.client.load_system_host_keys()

    #     # if self.auth_interactive_dumb:
    #     #     from lmn.helper.ssh import overwrite_auth_fn
    #     #     conn.client = overwrite_auth_fn(conn.client)

    #     self._conn = conn
    #     return conn

    def get_dict(self):
        return {key: val for key, val in vars(self).items() if not (key.startswith('__') or callable(val))}

    def __del__(self):
        # Close the connection when the instance is destructed
        if self._conn:
            self._conn.close()


class CLISSHClient:
    """Use the native CLI to run ssh"""
    def __init__(self, remote_conf: RemoteConfig) -> None:
        self.remote_conf = remote_conf
        # self.conn = self.remote_conf.get_connection()

    def uri(self, path):
        return f'{self.remote_conf.base_uri}:{path}'

    def run(self, cmd, directory="$HOME", env=None, capture_output: bool = False, dry_run: bool = False) -> Optional[str]:
        from lmn.cli._utils import run_cmd
        import re
        """
        Args:
            - capture_output (bool): If True, the returned object will contain stdout/stderr, **but the process won't be interactive**
        """
        # TODO: remove `env`: handle that within `cmd`!
        # TODO: Support disown (if that is necessary for sweep)
        env = {} if env is None else env

        # Perform shell escaping for envvars
        # NOTE: Fabric seems to have an issue to handle envvar that contains spaces...
        # This is the issue of using "inline_ssh_env" that essentially sets envvars by putting bunch of export KEY=VAL before running shells.
        # The documentation clearly says developers need to handle shell escaping for non-trivial values.
        # TEMP: shell escaping only when env contains space
        import shlex
        env = {key: shlex.quote(str(val)) if " " in str(val) else str(val) for key, val in env.items()}

        # NOTE:
        # -t: Force pseudo-terminal allocation.
        ssh_options = ['-t', f'-o ControlPath=~/.ssh/lmn-ssh-socket-{self.remote_conf.host}']
        ssh_base_cmd = f'ssh {" ".join(ssh_options)} {self.remote_conf.base_uri}'

        remote_cmds = []
        if directory is not None:
            remote_cmds += [f'cd {directory}']
            # ssh_cmd += f'cd {directory} && '
        remote_cmds += [cmd]

        ssh_cmd = f"{ssh_base_cmd} '{' && '.join(remote_cmds)}'"
        result = run_cmd(ssh_cmd, get_output=capture_output)
        return result

    def put(self, fpath, target_path=None) -> None:
        from lmn.cli._utils import run_cmd
        # TODO: Move the ControlPath to global config
        options = [f'-o ControlPath=$HOME/.ssh/lmn-ssh-socket-{self.remote_conf.host}']
        options = [os.path.expandvars(opt) for opt in options]
        cmd = ['scp', *options, fpath, f'{self.remote_conf.base_uri}:{target_path}']
        # Setting get_output=True has a side-effect of not showing stdout on the terminal
        run_cmd(cmd, shell=False, get_output=True)

    def port_forward(self):
        raise NotImplementedError

    def x_forward(self):
        raise NotImplementedError


# DEPRECATED
class SimpleSSHClient:
    """Given a remote config, this provides an interface to ssh into a remote machine.
    """
    def __init__(self, remote_conf: RemoteConfig) -> None:
        self.remote_conf = remote_conf
        self.conn = self.remote_conf.get_connection()

    def uri(self, path):
        return f'{self.remote_conf.base_uri}:{path}'

    def run(self, cmd, directory='$HOME', disown=False, hide=False, env=None, pty=False, dry_run=False):
        import re
        
        # TODO: Check if $HOME would work or not!!
        env = {} if env is None else env

        # Perform shell escaping for envvars
        # NOTE: Fabric seems to have an issue to handle envvar that contains spaces...
        # This is the issue of using "inline_ssh_env" that essentially sets envvars by putting bunch of export KEY=VAL before running shells.
        # The documentation clearly says developers need to handle shell escaping for non-trivial values.
        # TEMP: shell escaping only when env contains space
        import shlex
        env = {key: shlex.quote(str(val)) if " " in str(val) else str(val) for key, val in env.items()}

        if dry_run:
            logger.info('--- dry run ---')
            logger.info(f'cmd: {cmd}')
            logger.debug(locals())
        else:
            with self.conn.cd(directory):
                # promise = self.conn.run(cmd, asynchronous=True)
                if disown:
                    # NOTE: asynchronous=True --> disown=True
                    # asynchronous=True returns a Promise to which you can attach and listen to stdio.
                    # disown=True completely disowns the process.
                    self.conn.run(cmd, disown=True, hide=hide, env=env, pty=pty)
                    return

                # NOTE: if you use asynchronous=True, stdout/stderr does not show up
                # when you use it on slurm. I have no idea why, tho.
                logger.debug(f'ssh client env: {env}')
                try:
                    result = self.conn.run(cmd, asynchronous=False, hide=hide, env=env, pty=pty)
                except invoke.exceptions.UnexpectedExit as e:
                    logger.info(f'Caught an exception!!:\n{str(e)}')
                    import sys
                    sys.exit(1)

            return result

    def put(self, file_like, target_path=None):
        self.conn.put(file_like, str(target_path))

    def port_forward(self):
        raise NotImplementedError

    def x_forward(self):
        raise NotImplementedError



"""Load config file and fuse it with runtime options"""
from __future__ import annotations
from argparse import Namespace
from typing import Optional, List, Union
import os
import pathlib
from pathlib import Path
from lmn import logger
from lmn.helpers import find_project_root, parse_config
from posixpath import expandvars

from lmn.machine import RemoteConfig

DOCKER_ROOT_DIR = '/lmn'

# NOTE: I used to set it to /tmp/lmn, but that causes an issue:
# When multiple users use this directory, the one who used this first set the permission of /tmp/lmn
# to be his/hers, thus others trying to use it later cannot access it.
REMOTE_ROOT_DIR = '/tmp'

class Project:
    """Maintains the info specific to the local project"""
    def __init__(self, name: str, rootdir: Union[str, Path],
                 outdir: Optional[str] = None, exclude: Optional[List[str]] = None,
                 startup: Union[str, List[str]] = "",
                 mount_from_host: Optional[dict] = None,
                 env: Optional[dict] = None) -> None:
        self.name = name
        self.rootdir = Path(rootdir)
        self.outdir = self.rootdir / ".output" if outdir is None else outdir
        self.exclude = exclude
        if isinstance(startup, list):
            startup = '; '.join(startup)
        self.startup = startup
        self.env = env if env is not None else {}
        self.mount_from_host = mount_from_host if mount_from_host is not None else {}

        self._make_directories()

    def _make_directories(self):
        self.rootdir.mkdir(parents=True, exist_ok=True)
        self.outdir.mkdir(parents=True, exist_ok=True)

    def get_dict(self):
        return {key: val for key, val in vars(self).items() if not (key.startswith('__') or callable(val))}

    def __repr__(self):
        return repr(f'<Project {self.name}>')


class Machine:
    """Maintains machine configuration.
    - RemoteConfig (user, hostname, uri)
    """
    def __init__(self, remote_conf: RemoteConfig, lmndir: Union[str, Path],
                 parsed_conf: dict,
                 startup: Union[str, List[str]] = "",
                 env: Optional[dict] = None) -> None:
        self.remote_conf = remote_conf
        self.lmndir = Path(lmndir)
        self.env = env if env is not None else {}
        startup = startup or parsed_conf.startup
        if isinstance(startup, list):
            startup = '; '.join(startup)
        self.startup = startup
        self.parsed_conf = parsed_conf

        # aliases
        self.user = remote_conf.user
        self.host = remote_conf.host
        self.base_uri = remote_conf.base_uri

    def uri(self, path) -> str:
        """Returns user@hostname:path"""
        return f'{self.remote_conf.base_uri}:{path}'
    
    def get_lmndirs(self, project_name: str) -> Namespace:
        rootdir = self.lmndir / project_name
        return Namespace(
            codedir=str(rootdir / 'code'),
            mountdir=str(rootdir / 'mount'),
            outdir=str(rootdir / 'output'),
            scriptdir=str(rootdir / 'script'),
        )


def get_docker_lmndirs(lmndir: Union[Path, str], project_name: str) -> Namespace:
    rootdir = Path(lmndir) / project_name
    return Namespace(
        codedir=str(rootdir / 'code'),
        mountdir=str(rootdir / 'mount'),
        outdir=str(rootdir / 'output'),
        scriptdir=str(rootdir / 'script'),
    )


def load_config(machine_name: str):
    from lmn.config import ProjectConfig, MachineConfig
    proj_rootdir = find_project_root()
    config = parse_config(proj_rootdir)

    # Error checking in config file
    if 'machines' not in config:
        logger.error('The configuration file does not contain "machines" section.')
        import sys; sys.exit(1)

    if machine_name not in config['machines']:
        logger.error(
            f'Machine "{machine_name}" not found in your configuration. \n'
            f'Available machines are: {", ".join(config["machines"].keys())}'
        )
        import sys; sys.exit(1)

    # TODO:
    # Both pconf and mconf can definitely be pydantic objects

    _pconf = config.get('project', {})
    pconf = ProjectConfig(**_pconf)
    if pconf.name is None:
        pconf.name = proj_rootdir.stem

    mconf = MachineConfig(**config['machines'][machine_name])

    # Parse special config params
    preset_conf = {
        'slurm-configs': config.get('slurm-configs', {}),
        'pbs-configs': config.get('pbs-configs', {}),
        'docker-images': config.get('docker-images', {}),
    }

    logger.info(f'Project name     : {pconf.name}')
    logger.info(f'Project directory: {proj_rootdir}')

    # Load extra env vars from .env.secret
    secret_env_path = (proj_rootdir / ".secret.env").resolve()
    if not secret_env_path.is_file():
        secret_env_path = (proj_rootdir / ".env.secret").resolve()

        # Just a friendly reminder
        if secret_env_path.is_file():
            logger.info('Reading from ".env.secret" file will be deprecated in the future. Please rename it to ".secret.env".')

    from dotenv import dotenv_values
    secret_env = dotenv_values(secret_env_path)

    if secret_env:
        logger.debug(f'Loaded the following envs from secret env file: {list(dict(secret_env).keys())}')

    # TODO: Project should take ProjectConfig object as an argument
    project = Project(pconf.name,
                      proj_rootdir,
                      outdir=pconf.outdir,
                      exclude=pconf.exclude,
                      startup=pconf.startup,
                      env={**pconf.environment, **secret_env},
                      mount_from_host={**pconf.mount_from_host, **mconf.mount_from_host})

    remote_conf = RemoteConfig(mconf.user, mconf.host)

    if mconf.root_dir is None:
        lmndir = f'{REMOTE_ROOT_DIR}/{remote_conf.user}/lmn'
    else:
        lmndir = f"{mconf.root_dir}/{remote_conf.user}"
    machine = Machine(remote_conf,
                      parsed_conf=mconf,
                      lmndir=lmndir,
                      env=mconf.environment)

    return project, machine, preset_conf



#!/usr/bin/env python3
"""Simply run nvidia-smi command on a specified remote."""


from __future__ import annotations
from copy import deepcopy
import os
from pathlib import Path
from argparse import ArgumentParser
from argparse import Namespace
from lmn import logger
from lmn.cli.brun import handler as brun_handler

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from lmn.cli._config_loader import Project, Machine


def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "machine",
        action="store",
        type=str,
        help="Machine",
    )
    parser.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help="Be verbose"
    )
    return parser


def handler(project: Project, machine: Machine, parsed: Namespace, preset: dict):
    parsed.remote_command = 'nvidia-smi'
    brun_handler(project, machine, parsed, preset)


name = 'nv'
description = 'run nvidia-smi on a remote server'
parser = _get_parser()



from __future__ import annotations
from pydantic import BaseModel
from typing import Optional, List, Union

from lmn.container import DockerContainerConfig, SingularityConfig
from lmn.scheduler import SlurmConfig, PBSConfig


class ProjectConfig(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    environment: dict = {}
    outdir: Optional[str] = None
    mount_from_host: dict = {}
    exclude: List[str] = []
    startup: Union[str, List[str]] = ''


class MachineConfig(BaseModel):
    user: str
    host: str
    root_dir: Optional[str] = None
    environment: dict = {}
    mount_from_host: dict = {}  # Deprecated; Moved to container config (Docker and Singularity)
    startup: Union[str, List[str]] = ''
    mode: str = 'ssh'

    # Container config
    docker: Optional[DockerContainerConfig] = None
    singularity: Optional[SingularityConfig] = None

    # Scheduler config
    slurm: Optional[SlurmConfig] = None
    pbs: Optional[PBSConfig] = None


from pydantic import BaseModel, Field
from typing import Optional, List, Union
from lmn import logger


class SingularityConfig(BaseModel):
    sif_file: str
    env: dict = Field(alias="environment", default={})
    env_from_host: List[str] = []
    mount_from_host: dict = {}
    overlay: Optional[str] = None
    startup: Union[str, List[str], None] = None
    pwd: Optional[str] = None

    runtime: str = "singularity"
    writable_tmpfs: bool = True
    nv: bool = True
    containall: bool = True

    # HACK: I don't know how to appropriately set alias with Pydantic...
    @property
    def image(self):
        print("Image property is deprecated. Use sif_file instead.")
        return self.sif_file


class SingularityCommand:
    @staticmethod
    def _valid_key(key: str) -> str:
        '''Long arguments (for slurmCommand) constructed with '-' have been internally
        represented with '_' (for Python). Correct for this in the output.
        '''
        return key.replace('_', '-')

    @staticmethod
    def run(cmd: str, singularity_config: SingularityConfig) -> str:
        c = singularity_config
        options = SingularityCommand.make_options(c)

        if c.startup:
            startup = " ; ".join(c.startup) if isinstance(c.startup, list) else c.startup
            cmd = f"{startup} ; {cmd}"
        # Escape special chars and quotes (https://stackoverflow.com/a/18886646/19913466)
        logger.debug(f"cmd before escape: {cmd}")
        cmd = cmd.encode("unicode-escape").replace(b'"', b'\\"').decode("utf-8")
        logger.debug(f"cmd after escape: {cmd}")

        # NOTE: Without --containall, nvidia-smi command fails with "couldn't find libnvidia-ml.so library in your system."
        # NOTE: Without bash -c '{cmd}', if you put PYTHONPATH=/foo/bar, it fails with no such file or directory 'PYTHONPATH=/foo/bar'
        # TODO: Will the envvars be taken over to the internal shell (by this extra bash command)?
        return f'{c.runtime} run {" ".join(options)} {c.sif_file} bash -c -- "{cmd}"'

    @staticmethod
    def make_options(singularity_config: SingularityConfig) -> List[str]:
        c = singularity_config
        _valid_key = SingularityCommand._valid_key

        # Handle binary flags
        binary_flags = ["nv", "containall", "writable_tmpfs"]
        options = [f"--{_valid_key(flag)}" for flag in binary_flags if getattr(c, flag, False)]

        # Handle arg-val patterns (e.g., `--overlay /blah/blah.fs`)
        arg_vals = ["overlay", "pwd"]
        options += [
            f"--{_valid_key(arg)} {getattr(c, arg)}" for arg in arg_vals if getattr(c, arg, False)
        ]

        # Handle environment variables
        # TODO: Better to use --env-file option and read from a file
        # Escaping quotes and commas will be much easier in that way.
        if c.env:
            options += [
                "--env " + ",".join(f'{key}="{val}"' for key, val in c.env.items())
            ]

        if c.env_from_host:
            options += [
                "--env " + ",".join(f'{envvar}=${envvar}' for envvar in c.env_from_host)
            ]

        # Handle binds:
        bind = "-B {source}:{target}"
        options += [
            bind.format(source=source, target=target)
            for source, target in c.mount_from_host.items()
        ]

        return options


if __name__ == "__main__":
    sing_conf = SingularityConfig(
        sif_file="hoge.sif",
        overlay="foo/bar.fs",
        nv=True,
        containall=True,
        environment={"FOO": "BAR", "HOGE": "PIYO"},
        mount_from_host={"/foo/bar": "/foo/bar"},
    )
    sing_command = SingularityCommand.run("test-command", sing_conf)
    print(sing_command)



from pydantic import BaseModel, Field
from typing import Optional, List, Union


def get_gpu_device():
    """Add gpus to device_requests, just for dockerpy

    https://github.com/docker/docker-py/issues/2395#issuecomment-907243275
    """
    from docker.types import DeviceRequest
    gpu = DeviceRequest(
        count=-1,
        capabilities=[['gpu'], ['nvidia'], ['compute'], ['compat32'], ['graphics'], ['utility'], ['video'], ['display']]
    )

    return gpu


class DockerContainerConfig(BaseModel):
    image: str
    name: Optional[str] = None
    env: dict = Field(alias="environment", default={})
    mount_from_host: dict = {}
    remove: bool = True
    network: str = 'bridge'
    ipc_mode: str = 'private'
    startup: Union[str, List[str], None] = None
    tty: bool = True
    gpus: str = 'all'
    user_id: int = 0
    group_id: int = 0
    runtime: str = 'docker'


#!/usr/bin/env python3
from typing import List, Optional
from copy import deepcopy
from pydantic import BaseModel


class PBSConfig(BaseModel):
    job_name: str = 'default-job-name'  # To be filled
    account: str = 'SuperBERT'  # Account string
    queue: str = 'debug'  # debug, small, medium, large, etc. (Check `qstat -q`)
    shell: str = 'bash'  # It will be fed to `/bin/{shell}` template
    filesystems: str = 'home:grand'  # Request access to /home and /grand directories
    select: int = 1  # Request 1 node
    place: str = 'free'  # scatter, pack (default): specify how to distribute allocations (I believe it only matters for multi-node allocation ?)
    walltime: str = '1:00:00'  # 1 hour for debug queue, 72 hours for preemptable queue

    @property
    def resource_list(self):
        return [
            f'place={self.place}', f'walltime={self.walltime}',
            f'filesystems={self.filesystems}', f'select={self.select}'
        ]


class PBSCommand:
    @staticmethod
    def _valid_key(key: str) -> str:
        '''Long arguments (for slurmCommand) constructed with '-' have been internally
        represented with '_' (for Python). Correct for this in the output.
        '''
        return key.replace('_', '-')

    @staticmethod
    def make_options(pbs_config: dict) -> List[str]:
        """Return a list of arguments.
        Example: ['-I', '-l filesystems=home:grand', '-l select=1', ...]
        """
        needs_special_handling = ['l', 'I']
        pbs_config = deepcopy(pbs_config)

        # # Process space
        # pbs_config = {
        #     k: f"'{v}'" if isinstance(v, str) and " " in v else v
        #     for k, v in pbs_config.items()
        # }

        # We need a special handling on `-l` option, since it is allowed to appear multiple times in a command.
        args = [
            f"-{PBSCommand._valid_key(k)} {v}"
            for k, v in pbs_config.items()
            if k not in needs_special_handling and v is not None
        ]

        # Expand `-l` option (list)
        if 'l' in pbs_config:
            assert isinstance(pbs_config['l'], (list, tuple))
            args += [f'-l {v}' for v in pbs_config['l']]

        # Handle `-I` option (bool)
        if 'I' in pbs_config:
            assert isinstance(pbs_config['I'], bool)
            if pbs_config['I']:
                args += ['-I']

        return args

    @staticmethod
    def qsub_from_dict(run_cmd: str, pbs_config: dict, qsub_cmd: str = 'qsub', interactive: bool = False,
                       convert: bool = True):

        args = PBSCommand.make_options(pbs_config)
        if interactive:
            # NOTE:
            # ### Issues in interactive mode ###
            # It's tricky to execute a script in an interactive mode on PBS.
            # Whenever we use `-I` (interactive) option on qsub like:
            # `qsub <options> -I <script>`,
            # the script is simply ignored (actually, only PBS directives, the lines starting with `#PBS`, are respected).
            # ### Workaround ###
            # To overcome this issue, we can use `--` option with `bash -c` like:
            # `qsub <options> -I -- /usr/bin/env bash -c '<command>'`
            # or
            # `qsub <options> -I -- /bin/bash -c '<command>'`
            # This way, we can execute a script and a command in an interactive mode.
            # If we use a script, just make sure that `bash -c` is included in your script file.
            # ### Another workaround ###
            # I found that we can also use `-S` option to execute a script (this is originally for specifying a shell).
            # `qsub <options> -I -S <script>`
            cmd = ' '.join((qsub_cmd, *args, '--', run_cmd))
        else:
            batch_args = (f'#PBS {arg}' for arg in args)
            cmd = '\n'.join((
                qsub_cmd + ' << EOF',
                '\n'.join(('#!/usr/bin/env bash', '', *batch_args, '')),
                run_cmd.replace('$', '\\$') if convert else run_cmd,
                'EOF',
            ))

        return cmd

    @staticmethod
    def qsub(run_cmd: str, pbs_config: PBSConfig, qsub_cmd: str = 'qsub', interactive: bool = False,
             convert: bool = True):

        pbs_dict = {
            # a: # date and time
            'A': pbs_config.account,  # Account string
            # c
            # C
            # 'e': pbs_config.stderr,  # Path to be used for the job's standard error stream
            # 'f': pbs_config.force_foreground,  # Prevents qsub from spawning a background process.
            # h
            'I': interactive,  # Job is to be run interactively
            # 'j': # Specifies whether and how to join the job's standard error and standard output streams.
            # 'J': pbs_config.array,  # Makes this job an array job. format: `-J <range> [%<max subjobs>]`
            # 'k': # Specifies whether and which of the standard output and standard error streams is left behind on the execution host, or written to their final destination. (Default: ; neither is retained)
            'l': pbs_config.resource_list,  # Allows the user to request resources and specify job placement.
            # m
            # M
            'N': pbs_config.job_name,
            # 'o': pbs_config.stdout,  # Path to be used for the job's standard output stream.
            # 'p': # Priority of the job.  Sets job's Priority attribute to priority. (Range: [-1024, 1023], Default: Zero)
            # P
            'q': pbs_config.queue,  # Where the job is sent upon submission.
            # 'r':  # Declares whether the job is rerunnable. format: `-r <y|n>`
            # R  # Specifies whether standard output and/or standard error files are automatically removed (deleted) upon job completion.
            # 'S': pbs_config.shell,
            # u
            # 'v': ...,  # Lists environment variables and shell functions to be exported to the job. <- Let's rather use export explicitly.
            # V  # All environment variables and shell functions in the user's login environment where qsub is run are exported to the job.
            # 'W': ...,  # The -W option allows specification of some job attributes.  Some job attributes must be specified using this option.
            # z  # Job identifier is not written to standard output.
        }
        return PBSCommand.qsub_from_dict(run_cmd, pbs_dict, qsub_cmd, interactive, convert)



#!/usr/bin/env python3
from __future__ import annotations
from pydantic import BaseModel
from typing import Optional

# TODO: Implement SlurmCommand by myself
from simple_slurm_command import SlurmCommand


class SlurmConfig(BaseModel):
    """Let's keep it minimal. Conceptually it's better to store SlurmConfig here,
    but that would make it harder to read.
    """
    job_name: str = 'default-job-name'  # To be filled
    partition: str = 'cpu'
    constraint: Optional[str] = None
    cpus_per_task: int = 1
    time: Optional[str] = None
    output: Optional[str] = None
    error: Optional[str] = None
    dependency: Optional[str] = None
    nodelist: Optional[str] = None
    exclude: Optional[str] = None
    shell: str = 'bash'


#!/usr/bin/env python3
"""Bare-run: Run the given command without rsync, container setup, startup command or cd to workdir."""


from __future__ import annotations
from copy import deepcopy
import os
from pathlib import Path
from argparse import ArgumentParser
from argparse import Namespace
from lmn import logger
from lmn.helpers import find_project_root
from lmn.helpers import replace_lmn_envvars
from lmn.cli._config_loader import Project, Machine
from lmn.machine import CLISSHClient

from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ._config_loader import Machine

def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "machine",
        action="store",
        type=str,
        help="Machine",
    )
    parser.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help="Be verbose"
    )
    parser.add_argument(
        "remote_command",
        default=False,
        action="store",
        nargs="+",
        type=str,
        help="Command to execute in a remote machine.",
    )
    return parser


def print_conf(mode: str, machine: Machine, image: Optional[str] = None):
    output = f'Running with [{mode}] mode on [{machine.remote_conf.base_uri}]'
    if image is not None:
        output += f' with image: [{image}]'
    logger.info(output)

def handler(project: Project, machine: Machine, parsed: Namespace, preset: dict):
    logger.debug(f'handling command for {__file__}')
    logger.debug(f'parsed: {parsed}')

    # HACK
    mode = 'ssh'

    if isinstance(parsed.remote_command, list):
        cmd = ' '.join(parsed.remote_command)
    else:
        cmd = parsed.remote_command

    # Runtime info
    runtime_options = Namespace(dry_run=parsed.dry_run,
                                rel_workdir=None,
                                cmd=cmd)

    env = {**project.env, **machine.env}
    lmndirs = machine.get_lmndirs(project.name)

    startup = ' ; '.join([e for e in [project.startup, machine.startup] if e.strip()])

    # If parsed.mode is not set, try to read from the config file.
    from lmn.runner import SSHRunner
    ssh_client = CLISSHClient(machine.remote_conf)
    ssh_runner = SSHRunner(ssh_client, lmndirs)
    print_conf(mode, machine)
    ssh_runner.exec(runtime_options.cmd,
                    relative_workdir=runtime_options.rel_workdir,
                    startup=startup,
                    env=env,
                    dry_run=runtime_options.dry_run)


name = 'brun'
description = 'Bare run: Just run the command without any setup'
parser = _get_parser()



from __future__ import annotations
import time
from copy import deepcopy
import os
from pathlib import Path
from argparse import ArgumentParser
from argparse import Namespace
from lmn import logger
from lmn.helpers import find_project_root, replace_lmn_envvars
from lmn.machine import CLISSHClient
from lmn.runner import SlurmRunner, PBSRunner
from lmn.cli.sync import _sync_output, _sync_code
from lmn.const import available_modes


from typing import TYPE_CHECKING, Literal, Optional
if TYPE_CHECKING:
    from lmn.cli._config_loader import Project, Machine


def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "machine",
        action="store",
        type=str,
        help="Machine",
    )
    parser.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help="Be verbose"
    )
    parser.add_argument(
        "--image",
        default=None,
        help="specify a docker image"
    )
    parser.add_argument(
        "--name",
        default=None,
        help="specify docker container name"
    )
    parser.add_argument(
        "--sconf",
        default=None,
        help="specify a slurm configuration to be used"
    )
    parser.add_argument(
        "--pbsconf",
        default=None,
        help="specify a PBS configuration to be used"
    )
    parser.add_argument(
        "--dconf",
        default=None,
        help="specify a docker configuration to be used"
    )
    parser.add_argument(
        "-m",
        "--mode",
        action="store",
        type=str,
        default=None,
        choices=available_modes,
        help="What mode to run",
    )
    parser.add_argument(
        "-d",
        "--disown",
        action="store_true",
        help="Do not block to wait for the process to exit. stdout/stderr will not be shown with this option.",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="When a job with the same name already exists, kill it and run the new one (only for Docker mode)",
    )
    # parser.add_argument(
    #     "-X",
    #     "--x-forward",
    #     action="store_true",
    #     help="X11 forwarding",
    # )
    parser.add_argument(
        "--no-sync",
        action="store_true",
        help="Do not perform rsync. This means your local files will not be synced with remote server.",
    )
    parser.add_argument(
        "--contain",
        action="store_true",
        help="With this flag, rsync will copy the project directory to a new unique location on remote, rather than the predetermined one.",
    )
    parser.add_argument(
        "-n",
        "--num-sequence",
        action="store",
        type=int,
        default=1,
        help="number of sequence in Slurm sequential jobs"
    )
    parser.add_argument(
        "--sweep",
        action="store",
        type=str,
        help="specify sweep range (e.g., --sweep 0-255) this changes the value of $LMN_RUN_SWEEP_IDX"
    )
    parser.add_argument(
        "remote_command",
        default=False,
        action="store",
        nargs="+",
        type=str,
        help="Command to execute in a remote machine.",
    )
    return parser


def parse_sweep_idx(sweep_str):
    # Parse input
    # format #0: 8 --> 8
    # format #1: 1-10 --> range(1, 10)
    # format #2: 1,2,7 --> [1, 2, 7]
    if '-' in sweep_str:
        # format #1
        begin, end = [int(val) for val in sweep_str.split('-')]
        assert begin < end
        sweep_ind = range(begin, end)
    elif ',' in sweep_str:
        sweep_ind = [int(e) for e in sweep_str.strip().split(',')]
    elif sweep_str.isnumeric():
        sweep_ind = [int(sweep_str)]
    else:
        raise KeyError("Format for --sweep option is not recognizable. Format examples: '1-10', '8', '1,2,7'.")

    return sweep_ind


def print_conf(mode: str, machine: Machine, image: Optional[str] = None):
    output = f'Running with [{mode}] mode on [{machine.remote_conf.base_uri}]'
    if image is not None:
        output += f' with image: [{image}]'
    logger.info(output)


def handler(project: Project, machine: Machine, parsed: Namespace, preset: dict):
    """
    Args:
    - project (Project): stores project-specific configurations
    - machine (Machine): stores machine-specific configurations
    - preset (dict)    : stores preset configurations for slurm or docker images
    """
    logger.debug(f'handling command for {__file__}')
    logger.debug(f'parsed: {parsed}')

    # Runtime info
    curr_dir = Path(os.getcwd()).resolve()
    proj_rootdir = find_project_root()
    rel_workdir = curr_dir.relative_to(proj_rootdir)
    logger.debug(f'relative working dir: {rel_workdir}')  # cwd.relative_to(project_root)
    if isinstance(parsed.remote_command, list):
        cmd = ' '.join(parsed.remote_command)
    else:
        cmd = parsed.remote_command

    # TODO: Let's use Pydantic
    runtime_options = Namespace(dry_run=parsed.dry_run,
                                cmd=cmd,
                                rel_workdir=rel_workdir,
                                disown=parsed.disown,
                                name=parsed.name,
                                sweep=parsed.sweep,
                                num_sequence=parsed.num_sequence,
                                no_sync=parsed.no_sync,
                                sconf=parsed.sconf,
                                dconf=parsed.dconf,
                                force=parsed.force)

    # TODO:
    # - If configured, run a pre-flight ssh with ControlMaster to establish & retain the connection
    # - The following ssh / rsync should rely on this connection
    from lmn.helpers import establish_persistent_ssh
    establish_persistent_ssh(machine.remote_conf)

    # Sync code first
    if parsed.no_sync:
        logger.warning('--no-sync option is True, local files will not be synced.')

    if not parsed.no_sync:
        if parsed.contain:
            # Generate a unique path and set it to machine.lmndir
            # BUG: This generates the same hash every time!! This stack overflow answer is obviously wrong: https://stackoverflow.com/a/6048639/19913466
            # import hashlib
            # import time
            # hashlib.sha1().update(str(time.time()).encode("utf-8"))
            # _hash = hashlib.sha1().hexdigest()
            from lmn.helpers import get_timestamp
            _hash = get_timestamp()
            machine.lmndir = Path(f'{machine.lmndir}/{_hash}')
            runtime_options.name = _hash
            logger.info(f'--contain flag is set.\n\tsetting the remote lmndir to {machine.lmndir}\n\tsetting jobs suffix to {_hash}')

        _sync_code(project, machine, runtime_options.dry_run)


    # If parsed.mode is not set, try to read from the config file.
    mode = parsed.mode or machine.parsed_conf.mode
    if mode is None:
        logger.warning('mode is not set. Setting it to SSH mode')
        mode = 'ssh'

    if mode == 'ssh':
        from lmn.runner import SSHRunner
        # ssh_client = SimpleSSHClient(machine.remote_conf)

        startup = ' ; '.join([e for e in [project.startup, machine.startup] if e.strip()])
        lmndirs = machine.get_lmndirs(project.name)
        env = {**project.env, **machine.env}

        ssh_client = CLISSHClient(machine.remote_conf)
        ssh_runner = SSHRunner(ssh_client, lmndirs)
        print_conf(mode, machine)
        ssh_runner.exec(runtime_options.cmd,
                        runtime_options.rel_workdir,
                        startup=startup,
                        env=env,
                        dry_run=runtime_options.dry_run)

    elif mode == 'docker':
        from docker import DockerClient
        from lmn.runner import DockerRunner
        from lmn.container.docker import DockerContainerConfig

        # if 'docker' not in machine.parsed_conf:
        if machine.parsed_conf.docker is None:
            raise ValueError('Configuration must have an entry for "docker" to use docker mode.')

        startup = ' ; '.join([e for e in [project.startup, machine.startup] if e.strip()])
        lmndirs = machine.get_lmndirs(project.name)
        env = {**project.env, **machine.env}
        env_from_host = []

        base_url = "ssh://" + machine.base_uri
        # client = DockerClient(base_url=base_url, use_ssh_client=True)
        client = DockerClient(base_url=base_url)  # dockerpty hangs with use_ssh_client=True

        # Specify job name
        name = f'{machine.user}-lmn-{project.name}'
        if runtime_options.name is not None:
            name = f'{name}--{runtime_options.name}'

        if runtime_options.dry_run:
            raise ValueError('dry run is not yet supported for Docker mode')

        from docker.types import Mount
        from lmn.cli._config_loader import DOCKER_ROOT_DIR, get_docker_lmndirs
        docker_lmndirs = get_docker_lmndirs(DOCKER_ROOT_DIR, project.name)

        # Load Docker-specific configurations
        # docker_pconf = DockerContainerConfig(**{
        #     'name': name,
        #     **machine.parsed_conf['docker']
        # })
        docker_pconf = machine.parsed_conf.docker
        docker_pconf.name = name

        if startup:
            logger.warn('`startup` configurations outside of `docker` will be ignored in docker mode.')
            logger.warn('Please place `startup` under `docker` if you want to run it in the container.')

        if not runtime_options.no_sync:
            docker_pconf.mount_from_host.update({
                f'{lmndirs.codedir}': docker_lmndirs.codedir,
                f'{lmndirs.outdir}': docker_lmndirs.outdir,
                f'{lmndirs.mountdir}': docker_lmndirs.mountdir,
                f'{lmndirs.scriptdir}': docker_lmndirs.scriptdir,
            })

        if project.mount_from_host:
            # Backward compatible
            logger.warn("`mount_from_host` under `project` is deprecated. Please move it under `docker`.")
            docker_pconf.mount_from_host.update(project.mount_from_host)

        docker_pconf.env.update(env)

        docker_runner = DockerRunner(client, docker_lmndirs)

        print_conf(mode, machine, docker_pconf.image)
        if runtime_options.sweep:
            if not runtime_options.disown:
                logger.error("You must set -d option to use sweep functionality.")
                import sys; sys.exit(1)
            sweep_ind = parse_sweep_idx(runtime_options.sweep)

            single_sweep = (len(sweep_ind) == 1)

            for sweep_idx in sweep_ind:
                _name = f'{name}-{sweep_idx}'
                logger.info(f'Launching sweep {sweep_idx}: {_name}')
                env.update({'LMN_RUN_SWEEP_IDX': sweep_idx})

                dconf = deepcopy(docker_pconf)
                dconf.name = _name
                dconf.env.update(env)

                docker_runner.exec(runtime_options.cmd,
                                   runtime_options.rel_workdir,
                                   dconf,
                                   interactive=False,
                                   kill_existing_container=runtime_options.force,
                                   quiet=not single_sweep)
        else:
            docker_runner.exec(runtime_options.cmd,
                               runtime_options.rel_workdir,
                               docker_pconf,
                               # startup=startup,
                               interactive=not runtime_options.disown,
                               kill_existing_container=runtime_options.force)

    elif 'slurm' in mode or 'pbs' in mode:
        # validate the mode
        if mode not in ['slurm', 'slurm-sing', 'sing-slurm', 'pbs', 'pbs-sing', 'sing-pbs']:
            logger.error(f'Invalid mode: {mode}')
            import sys; sys.exit(1)

        handler_scheduler(project, machine, parsed, preset, mode, runtime_options)

    else:
        raise ValueError(f'Unrecognized mode: {mode}')

    # Sync output files
    if not runtime_options.no_sync and not runtime_options.disown:
        _sync_output(project, machine, dry_run=parsed.dry_run)


def handler_scheduler(
    project: Project,
    machine: Machine,
    parsed: Namespace,
    preset: dict,
    mode: Literal['slurm', 'slurm-sing', 'sing-slurm', 'pbs', 'pbs-sing', 'sing-pbs'],
    run_opt: Namespace,
    ):
    import random
    import randomname

    startup = ' ; '.join([e for e in [project.startup, machine.startup] if e.strip()])
    lmndirs = machine.get_lmndirs(project.name)
    env = {**project.env, **machine.env}

    ssh_client = CLISSHClient(machine.remote_conf)

    if 'slurm' in mode:
        from lmn.scheduler.slurm import SlurmConfig

        if machine.parsed_conf.slurm is None:
            raise ValueError('Configuration must have an entry for "slurm" to use Slurm mode.')

        # Create SlurmConfig object
        if parsed.sconf is None:
            scheduler_conf = machine.parsed_conf.slurm
        else:
            # Try to load from the slurm conf presets
            logger.debug('parsed.sconf is specified. Loading custom preset conf.')
            _sconf = preset.get('slurm-configs', {}).get(parsed.sconf, {})
            if not _sconf:
                logger.error(f'Slurm config preset: "{parsed.sconf}" cannot be found in "slurm-configs" or is empty.')
                import sys; sys.exit(1)
            scheduler_conf = SlurmConfig(**_sconf)
            logger.info(f'Using Slurm preset: [{parsed.sconf}]')

        runner = SlurmRunner(ssh_client, lmndirs)

    elif 'pbs' in mode:
        from lmn.scheduler.pbs import PBSConfig

        if machine.parsed_conf.pbs is None:
            raise ValueError('Configuration must have an entry for "pbs" to use PBS mode.')

        # Create PBSConfig object
        if parsed.pbsconf is None:
            scheduler_conf = machine.parsed_conf.pbs
        else:
            # Try to load from the slurm conf presets
            logger.debug('parsed.sconf is specified. Loading custom preset conf.')
            _pconf = preset.get('pbs-configs', {}).get(parsed.pbsconf, {})
            if not _pconf:
                logger.error(f'PBS config preset: "{parsed.pbsconf}" cannot be found in "pbs-configs" or is empty.')
                import sys; sys.exit(1)
            scheduler_conf = PBSConfig(**_pconf)
            logger.info(f'Using PBS preset: [{parsed.pbsconf}]')

        runner = PBSRunner(ssh_client, lmndirs)

    else:
        logger.error(f'Unrecognized mode: {mode}')
        import sys; sys.exit(1)

    # NOTE: Slurm / PBS seem to be fine with duplicated name.
    proj_name_maxlen = 15
    rand_num = random.randint(0, 100)
    job_name = f'lmn-{project.name[:proj_name_maxlen]}-{randomname.get_name()}-{rand_num}'

    # Overwrite job name
    scheduler_conf.job_name = job_name

    if run_opt.force:
        logger.warn("`-f / --force` option has no effect in Slurm / PBS mode")

    # Specify job name
    name = f'{machine.user}-lmn-{project.name}'
    if run_opt.name is not None:
        name = f'{name}--{run_opt.name}'
    scheduler_conf.job_name = name

    if 'sing' in mode:
        from lmn.container.singularity import SingularityConfig, SingularityCommand

        # Error checking
        if machine.parsed_conf.singularity is None:
            raise ValueError('Entry "singularity" not found in the config file.')

        # Overwrite lmn envvars.  Hmm I don't like this...
        from lmn.cli._config_loader import DOCKER_ROOT_DIR, get_docker_lmndirs
        sing_lmndirs = get_docker_lmndirs(DOCKER_ROOT_DIR, project.name)

        # Update `pwd` entry
        sing_conf = machine.parsed_conf.singularity
        sing_conf.pwd = str(sing_lmndirs.codedir / run_opt.rel_workdir)

        print('========= the value of env at this point is =======')
        print(env)

        # TODO: Use get_lmndirs function!
        # TODO: Do we need this??
        env.update({
            'LMN_PROJECT_NAME': project.name,
            'LMN_CODE_DIR': sing_lmndirs.codedir,
            'LMN_MOUNT_DIR': sing_lmndirs.mountdir,
            'LMN_OUTPUT_DIR': sing_lmndirs.outdir,
            'LMN_SCRIPT_DIR': sing_lmndirs.scriptdir,
        })

        # Backward compat
        env.update({
            'RMX_PROJECT_NAME': project.name,
            'RMX_CODE_DIR': sing_lmndirs.codedir,
            'RMX_MOUNT_DIR': sing_lmndirs.mountdir,
            'RMX_OUTPUT_DIR': sing_lmndirs.outdir,
            'RMX_SCRIPT_DIR': sing_lmndirs.scriptdir,
        })

        # Bind
        mount_from_host = {}
        if not run_opt.no_sync:
            mount_from_host.update({
                f'{lmndirs.codedir}': sing_lmndirs.codedir,
                f'{lmndirs.outdir}': sing_lmndirs.outdir,
                f'{lmndirs.mountdir}': sing_lmndirs.mountdir,
                f'{lmndirs.scriptdir}': sing_lmndirs.scriptdir,
            })
        if project.mount_from_host:
            # Backward compatible
            logger.warn("`mount_from_host` under `project` is deprecated. Please move it under `singularity`.")
            mount_from_host.update(project.mount_from_host)

        # NOTE: Since CUDA_VISIBLE_DEVICES is often comma-separated values (i.e., CUDA_VISIBLE_DEVICES=0,1),
        # and `singularity run --env FOO=BAR,HOGE=PIYO` considers comma to be a separator for envvars,
        # It fails without special handling.
        sing_conf.env_from_host += ['SLURM_JOBID', 'SLURM_JOB_ID', 'SLURM_TASK_PID']

        # Integrate current `env`, `mount_from_host`
        sing_conf.env.update(env)
        sing_conf.mount_from_host.update(mount_from_host)

        # Finally overwrite run_opt.cmd
        run_opt.cmd = SingularityCommand.run(run_opt.cmd, sing_conf)

    timestamp = f"{time.time():.5f}".split('.')[-1]  # 5 subdigits of the timestamp
    print_conf(mode, machine, image=sing_conf.sif_file if mode in ['slurm-sing', 'sing-slurm'] else None)
    if run_opt.sweep:
        if not run_opt.disown:
            logger.error("You must set -d option to use sweep functionality.")
            import sys; sys.exit(1)

        if not parsed.contain:
            logger.error("You should set --contain option to use sweep functionality.")
            import sys; sys.exit(1)

        sweep_ind = parse_sweep_idx(run_opt.sweep)

        _scheduler_conf = deepcopy(scheduler_conf)
        for sweep_idx in sweep_ind:
            # NOTE: This special prefix "SINGULARITYENV_" is stripped and the rest is passed to singularity container,
            # even with --containall or --cleanenv !!
            # Example: (https://docs.sylabs.io/guides/3.1/user-guide/environment_and_metadata.html?highlight=environment%20variable)
            #     $ SINGULARITYENV_HELLO=world singularity exec centos7.img env | grep HELLO
            #     HELLO=world
            env.update({
                'SINGULARITYENV_LMN_RUN_SWEEP_IDX': sweep_idx,
                'APPTAINERENV_LMN_RUN_SWEEP_IDX': sweep_idx,
            })

            # Oftentimes, a user specifies $LMN_RUN_SWEEP_IDX as an argument to the command,
            # and that will be evaluated right before singularity launches
            env.update({'LMN_RUN_SWEEP_IDX': sweep_idx, 'RMX_RUN_SWEEP_IDX': sweep_idx})

            # Add sweep_idx to the job name
            _scheduler_conf.job_name = f'{scheduler_conf.job_name}-{timestamp}-{sweep_idx}'
            logger.info(f'Launching sweep {sweep_idx}: {_scheduler_conf.job_name}')

            runner.exec(run_opt.cmd, run_opt.rel_workdir, conf=_scheduler_conf,
                        startup=startup,
                        timestamp=timestamp,
                        interactive=False, num_sequence=run_opt.num_sequence,
                        env=env, dry_run=run_opt.dry_run)
    else:
        runner.exec(run_opt.cmd, run_opt.rel_workdir, conf=scheduler_conf,
                    startup=startup, timestamp=timestamp, interactive=not run_opt.disown, num_sequence=run_opt.num_sequence,
                    env=env, dry_run=run_opt.dry_run)


name = 'run'
description = 'run command'
parser = _get_parser()



#!/usr/bin/env python3
from __future__ import annotations
import os
import subprocess
import sys
from lmn import logger
from lmn.machine import RemoteConfig
from typing import List, Optional, Union
from pathlib import Path


def rsync(source_dir: Union[Path, str], target_dir: Union[Path, str], remote_conf: RemoteConfig, options: Optional[List[str]] = None,
          exclude: Optional[List[str]] = None, dry_run: bool = False, transfer_rootdir: bool = True, to_local: bool = False):
    """
    source_dir: hoge/fuga/source-dir/content-files
    target_dir: Hoge/Fuga/target-dir

    if transfer_rootdir is True:
      target_dir: Hoge/Fuga/target-dir/source-dir/content-files

    else:
      target_dir: Hoge/Fuga/target-dir/content-files
    """
    import shutil
    exclude = [] if exclude is None else exclude
    options = [] if options is None else options

    # make sure rsync is installed
    if shutil.which("rsync") is None:
        raise RuntimeError("rsync command is not found.")

    source_dir = str(source_dir).rstrip('/') + ('' if transfer_rootdir else '/')
    target_dir = str(target_dir).rstrip('/') + '/'

    # TODO: Move the ControlPath to global config
    options += [f'-e "ssh -o \'ControlPath=~/.ssh/lmn-ssh-socket-{remote_conf.host}\'"']
    options += ['--archive', '--compress']
    options += [f'--exclude \'{ex}\'' for ex in exclude]
    options_str = ' '.join(options)
    if to_local:
        cmd = f"rsync {options_str} {remote_conf.base_uri}:{source_dir} {target_dir}"
        logger.info(f"Syncing files ({remote_conf.base_uri}:{source_dir} to {target_dir})")
    else:
        cmd = f"rsync {options_str} {source_dir} {remote_conf.base_uri}:{target_dir}"
        logger.info(f"Syncing files ({source_dir} to {remote_conf.base_uri}:{target_dir})")

    if not dry_run:
        run_cmd(cmd, shell=True)
        logger.info("Sync finished!")


def run_cmd(cmd, get_output: bool = False, shell: bool = True, ignore_error: bool = False) -> Optional[str]:
    # TODO: 
    # - Do we ever need shell = False ??
    # - Do we ever need ignore_error = True ??
    # What is the downside of using get_output = True ??
    # - I guess you can't interact with the process with get_output = True.

    subprocess_env = dict(os.environ)
    logger.debug(f'running command: {cmd}')
    if get_output:
        result = subprocess.run(cmd, shell=shell, capture_output=True)
        if result.returncode != 0 and not ignore_error:
            stderr = result.stderr.read().decode('utf-8')
            msg = f"The command {cmd} returned exit code {result.returncode}\n---\n{stderr}\n---"
            raise RuntimeError(msg)
        return result.stdout.decode('utf-8').rstrip()

    result = subprocess.run(cmd,
                            stdout=sys.stdout, stderr=sys.stderr,
                            env=subprocess_env,
                            shell=shell)
    # Check for errors
    if not ignore_error and result.returncode != 0:
        msg = f"The command {cmd} returned exit code {result.returncode}"
        raise RuntimeError(msg)



from __future__ import annotations
import threading
import time
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, List, Optional, Union
from tempfile import NamedTemporaryFile
from lmn import logger
from lmn.helpers import replace_lmn_envvars


if TYPE_CHECKING:
    from argparse import Namespace
    from docker import DockerClient
    from lmn.machine import CLISSHClient
    from lmn.container.docker import DockerContainerConfig
    from lmn.scheduler.pbs import PBSConfig


def get_lmnenvs(cmd: str, lmndirs: Namespace):
    envvars = {
        "LMN_CODE_DIR": lmndirs.codedir,
        "LMN_MOUNT_DIR": lmndirs.mountdir,
        "LMN_OUTPUT_DIR": lmndirs.outdir,
        "LMN_SCRIPT_DIR": lmndirs.scriptdir,
        # "LMN_USER_COMMAND": cmd,
    }

    # Provide RMX_* envvars for backward compatibility
    return {
        **envvars,
        **{key.replace("LMN", "RMX"): val for key, val in envvars.items()},
    }


class SSHRunner:
    def __init__(self, client: CLISSHClient, lmndirs: Namespace) -> None:
        self.client = client
        self.lmndirs = lmndirs

    def exec(self, cmd: str, relative_workdir, env: Optional[dict] = None, startup: Union[str, List[str]] = "", dry_run: bool = False):
        env = {} if env is None else env
        # if isinstance(cmd, list):
        #     cmd = ' '.join(cmd) if len(cmd) > 1 else cmd[0]

        if startup:
            if isinstance(startup, list):
                startup = ' ; '.join(startup)
            cmd = f'{startup} ; {cmd}'

        logger.debug(f'ssh run with command: {cmd}')
        if relative_workdir is not None:
            workdir = self.lmndirs.codedir / relative_workdir
            logger.debug(f'cd to {workdir}')
        else:
            workdir = None
        lmnenv = get_lmnenvs(cmd, self.lmndirs)
        allenv = {**env, **lmnenv}
        allenv = {key: replace_lmn_envvars(val, lmnenv) for key, val in allenv.items()}
        return self.client.run(cmd, directory=workdir, env=allenv, dry_run=dry_run)


class DockerRunner:
    def __init__(self, client: DockerClient, lmndirs: Namespace) -> None:
        self.client = client
        self.lmndirs = lmndirs

    def exec(self, cmd: str, relative_workdir, docker_conf: DockerContainerConfig, startup: str = "",
             kill_existing_container: bool = True, interactive: bool = True, quiet: bool = False,
             log_stderr_background: bool = False, use_cli: bool = True) -> None:

        if startup:
            raise NotImplementedError(
                "Currently startup command before launching the container is not supported.\n"
                "For now, you can only specify `startup` command that runs in the container."
            )

        if log_stderr_background:
            assert not interactive, 'log_stderr_background=True cannot be used with interactive=True'

        lmnenv = get_lmnenvs(cmd, self.lmndirs)
        allenv = {**docker_conf.env, **lmnenv}
        allenv = {key: replace_lmn_envvars(val, lmnenv) for key, val in allenv.items()}

        # NOTE: target: container, source: remote host
        logger.debug(f'mounts: {docker_conf.mount_from_host}')
        logger.debug(f'docker_conf: {docker_conf}')

        logger.debug(f'container codedir: {str(self.lmndirs.codedir / relative_workdir)}')

        if kill_existing_container:
            from docker.errors import NotFound
            try:
                container = self.client.containers.get(docker_conf.name)
            except NotFound:
                container = None

            if container:
                logger.warning(f'Removing the existing container: {docker_conf.name}')
                container.remove(force=True)

        # NOTE: Consider using Volume rather than Mount
        # - https://docker-py.readthedocs.io/en/1.2.3/volumes/
        # - https://docs.docker.com/storage/bind-mounts/

        # NOTE: Intentionally being super verbose to make arguments explicit.
        d = docker_conf

        # HACK: if docker.startup is a list, flatten it to a string
        if isinstance(docker_conf.startup, list):
            docker_conf.startup = ' ; '.join(docker_conf.startup)

        # TEMP: When running in non-interactive mode and command fails, container disappears before we attach to its log stream,
        # and thus we cannot observe its error message. A naive way to avoid it is to wait for a bit before command execution.
        if not interactive:
            if docker_conf.startup:
                docker_conf.startup = ' ; '.join((docker_conf.startup, 'sleep 2'))
            else:
                docker_conf.startup = 'sleep 2'

        if docker_conf.startup:
            cmd = ' ; '.join((docker_conf.startup, cmd))
        cmd = f'{cmd} && chmod -R a+r {str(self.lmndirs.outdir)}'

        assert d.tty

        # Use python-on-whales (i.e., docker cli)
        import python_on_whales
        whale_client = python_on_whales.DockerClient(host=f'ssh://{self.client.api._custom_adapter.ssh_host}')
        logger.debug(f'docker run with command: {cmd}')

        if interactive:
            try:
                whale_client.run(
                    d.image,
                    ['/bin/bash', '-c', cmd],
                    detach=False,
                    envs=allenv,
                    gpus=d.gpus,
                    interactive=interactive,
                    ipc=d.ipc_mode,
                    mounts=[('type=bind', f'source={src}', f'destination={tgt}') for src, tgt in d.mount_from_host.items()],
                    name=d.name,
                    networks=[d.network],
                    remove=True,
                    tty=True,
                    user=f'{d.user_id}:{d.group_id}',
                    workdir=str(self.lmndirs.codedir / relative_workdir),
                )
            except python_on_whales.exceptions.DockerException as e:
                # NOTE: hide error as the exception is also raised when the command in the container returns non-zero exit value.
                import traceback
                logger.debug(f'python_on_whales failed!!:\n{str(e)}')
                logger.debug(traceback.format_exc())

        else:
            # NOTE:
            # Ideally, I'd like to use python-on-whales for non-interactive mode as well,
            # but somehow its detached mode runs very slowly...

            # Handle mount here
            from docker.types import Mount
            from lmn.container.docker import get_gpu_device
            mounts = [Mount(target=tgt, source=src, type='bind') for src, tgt in docker_conf.mount_from_host.items()]

            cmd = f'/bin/bash -c \'{cmd}\''
            container = self.client.containers.run(
                d.image,
                cmd,
                name=d.name,
                remove=d.remove,  # Keep it running as we need to change
                network=d.network,
                ipc_mode=d.ipc_mode,
                detach=True,
                tty=(not log_stderr_background),  # If True, stdout/stderr are mixed, but I observed sometimes some stdout are not showing up when tty=False
                stdin_open=True,  # It's useful to keep it open, as you may manually attach the container later
                mounts=mounts,
                environment=allenv,
                device_requests=[get_gpu_device()],
                working_dir=str(self.lmndirs.codedir / relative_workdir),
                user=f'{d.user_id}:{d.group_id}',
            )
            logger.debug(f'container: {container}')

            def log_stream(stream: Iterable):
                """Print out log stream."""
                for char in stream:
                    # logger.info(char.decode('utf-8'))
                    # 'ignore' ignores decode error that happens when multi-byte char is passed.
                    print(char.decode('utf-8', 'ignore'), end='')

            if not quiet:
                if log_stderr_background:
                    # Attach log stream in a separate thread, and only output stderr
                    stream = container.logs(stdout=False, stderr=True, stream=True, follow=True)
                    logger.info('quiet is True, only listening to stderr.')
                    logger.info('--- listening container stderr ---\n')
                    thr = threading.Thread(target=log_stream, args=(stream, ))
                    thr.start()
                    # thr.join()  # This blocks forever ;P

                else:
                    # Block and listen to the stream from container
                    stream = container.logs(stream=True, follow=True)
                    logger.info('--- listening container stdout/stderr ---\n')
                    log_stream(stream)


class SlurmRunner:
    """Use srun/sbatch to submit the command on a remote machine.
    If your local machine has slurm (i.e., you're on slurm login-node), I guess you don't need this tool.
    Thus SlurmMachine inherits SSHMachine.
    """
    def __init__(self, client: CLISSHClient, lmndirs: Namespace) -> None:
        self.client = client
        self.lmndirs = lmndirs

    def exec(self, cmd: str, relative_workdir, conf, env: Optional[dict] = None,
             env_from_host: List[str] = [],
             startup: Union[str, List[str]] = "", timestamp: str = "", num_sequence: int = 1,
             interactive: bool = None, dry_run: bool = False):
        """
        Args:
            - env_from_host (List[str]): Used for Singularity, inherit specified envvars from host
        """
        from lmn.scheduler.slurm import SlurmCommand
        env = {} if env is None else env

        # TODO: Verify env_from_list contains valid environment variable names
        # (i.e., cannot have spaces or quotes!!)

        if isinstance(cmd, list):
            cmd = ' '.join(cmd)

        if num_sequence > 1:
            # Use sbatch and set dependency to singleton
            conf.dependency = 'singleton'
            if interactive:
                logger.warning(f'num_sequence is set to {num_sequence} (> 1). Force disabling interactive mode')
                interactive = False

        s = conf
        slurm_command = SlurmCommand(cpus_per_task=s.cpus_per_task,
                                     job_name=s.job_name,
                                     partition=s.partition,
                                     time=s.time,
                                     nodelist=s.nodelist,
                                     exclude=s.exclude,
                                     constraint=s.constraint,
                                     dependency=s.dependency,
                                     output=s.output,
                                     error=s.error)

        if interactive and (s.output is not None):
            # User may expect stdout shown on the console.
            # logger.info('--output/--error argument for Slurm is ignored in interactive mode.')
            pass

        lmnenv = get_lmnenvs(cmd, self.lmndirs)
        allenv = {**env, **lmnenv}
        allenv = {key: replace_lmn_envvars(val, lmnenv) for key, val in allenv.items()}

        if startup:
            if isinstance(startup, list):
                startup = ' ; '.join(startup)
            cmd = f'{startup} ; {cmd}'

        workdir = Path(self.lmndirs.codedir) / relative_workdir

        slurm_options = []
        import shlex
        exports = [f'export {key}={shlex.quote(str(val))}' for key, val in allenv.items()]

        # DEPRECATED
        if env_from_host:
            logger.warn('"env_from_host" in "slurm" or "pbs" is deprecated. Please move it inside "singularity".')
            # Only matters for Singularity
            exports += [
                *[f'export SINGULARITYENV_{envvar}=${envvar}' for envvar in env_from_host],
                *[f'export APPTAINERENV_{envvar}=${envvar}' for envvar in env_from_host],
            ]

        if not interactive:
            sbatch_cmd = slurm_command.sbatch(cmd, shell=f'/usr/bin/env {s.shell}')
            # HACK: rather than submitting with `sbatch << EOF\n ...\n EOF`,
            # I use `sbatch file-name` to avoid `$ENVVAR` to be evaluated right at the submission time
            # I want the `$ENVVAR` to be evaluated after the compute is allocated.
            slurm_options += sbatch_cmd.split('\n')[2:-2]  # Strip `sbatch << EOF`, '#!/usr/bin/env/ bash', {cmd} and `EOF`

        exec_str = '\n'.join((
            # NOTE: without `-S` option, `bash -i` will be considered a single command and will end up in command not found.
            # Reference: https://unix.stackexchange.com/a/657774/556831
            # f'#!/usr/bin/env -S {s.shell} -i',
            f'#!/bin/{s.shell} -i',  # ^ This only works with a recent version of coreutils (8.30) and that cannot be taken for granted.
            *slurm_options,
            *exports,
            cmd
        ))
        logger.debug(f'\n=== execution string ===\n{exec_str}\n========================')

        script_fpath = Path(self.lmndirs.scriptdir) / f'.script-{timestamp}.sh'
        with NamedTemporaryFile(mode='w+') as temp_file:
            temp_file.write(exec_str)  # Write the string to the temporary file
            temp_file.flush()  # This is necessary!!
            self.client.put(temp_file.name, script_fpath)

        if interactive:
            cmd = slurm_command.srun(str(script_fpath), pty=s.shell)
        else:
            cmd = '\n'.join([f'sbatch {script_fpath}'] * num_sequence)

        try:
            self.client.run(cmd, directory=workdir, dry_run=dry_run)
        except RuntimeError as e:
            # NOTE: hide error as the exception is also raised when the command in the container returns non-zero exit value.
            import traceback
            logger.debug(f'self.client.run(...) failed!!:\n{str(e)}')
            logger.debug(traceback.format_exc())


class PBSRunner:
    def __init__(self, client: CLISSHClient, lmndirs: Namespace) -> None:
        self.client = client
        self.lmndirs = lmndirs

    def exec(self, cmd: str, relative_workdir, conf: PBSConfig, env: Optional[dict] = None,
             env_from_host: List[str] = [],
             startup: Union[str, List[str]] = "", timestamp: str = "", num_sequence: int = 1,
             interactive: bool = None, dry_run: bool = False):
        from lmn.scheduler.pbs import PBSCommand
        env = {} if env is None else env

        assert num_sequence == 1, 'Jobs with dependency are not supported on PBS mode yet.'

        if isinstance(cmd, list):
            cmd = ' '.join(cmd)

        lmnenv = get_lmnenvs(cmd, self.lmndirs)
        allenv = {**env, **lmnenv}
        allenv = {key: replace_lmn_envvars(val, lmnenv) for key, val in allenv.items()}

        if startup:
            # Use ';' rather than '&&' to avoid error when startup fails
            if isinstance(startup, list):
                startup = ' ; '.join(startup)
            cmd = f'{startup} ; {cmd}'

        workdir = Path(self.lmndirs.codedir) / relative_workdir

        slurm_options = []
        import shlex
        exports = [f'export {key}={shlex.quote(str(val))}' for key, val in allenv.items()]

        # DEPRECATED
        if env_from_host:
            logger.warn('"env_from_host" in "slurm" or "pbs" is deprecated. Please move it inside "singularity".')
            # Only matters for Singularity
            exports += [
                *[f'export SINGULARITYENV_{envvar}=${envvar}' for envvar in env_from_host],
                *[f'export APPTAINERENV_{envvar}=${envvar}' for envvar in env_from_host],
            ]

        if not interactive:
            qsub_cmd = PBSCommand.qsub(cmd, conf, interactive=False)
            # HACK: rather than submitting with `sbatch << EOF\n ...\n EOF`,
            # I use `sbatch file-name` to avoid `$ENVVAR` to be evaluated right at the submission time
            # I want the `$ENVVAR` to be evaluated after the compute is allocated.
            slurm_options += qsub_cmd.split('\n')[1:-2]  # Strip `qsub << EOF`, {cmd} and `EOF`

        exec_str = '\n'.join((
            # NOTE: without `-S` option, `bash -i` will be considered a single command and will end up in command not found.
            # Reference: https://unix.stackexchange.com/a/657774/556831
            # '#!/usr/bin/env -S bash -i' if interactive else '#!/usr/bin/env bash',
            f'#!/bin/{conf.shell} -i',  # ^ This only works with a recent version of coreutils (8.30) and that cannot be taken for granted.
            *slurm_options,
            *exports,
            cmd
        ))
        logger.debug(f'\n=== execution string ===\n{exec_str}\n========================')

        script_fpath = Path(self.lmndirs.scriptdir) / f'.script-{timestamp}.sh'
        with NamedTemporaryFile(mode='w+') as temp_file:
            temp_file.write(exec_str)  # Write the string to the temporary file
            temp_file.flush()  # This is necessary!!
            self.client.put(temp_file.name, script_fpath)

        if interactive:
            cmd = PBSCommand.qsub(str(script_fpath),
                                  conf,
                                  qsub_cmd=f'chmod +x {script_fpath} && qsub',  # HACK to make the script executable
                                  interactive=True)
        else:
            cmd = '\n'.join([f'qsub {script_fpath}'] * num_sequence)

        logger.debug(f'submission command: {cmd}')

        try:
            self.client.run(cmd, directory=workdir, dry_run=dry_run)
        except RuntimeError as e:
            # NOTE: hide error as the exception is also raised when the command in the container returns non-zero exit value.
            import traceback
            logger.debug(f'self.client.run(...) failed!!:\n{str(e)}')
            logger.debug(traceback.format_exc())